@@ -18,6 +18,9 @@ osmo-clean-workspace.sh
 
 mkdir "$deps" || true
 
+if [ "x$GTP" == "x--enable-gtp-linux" ]; then
+	osmo-build-dep.sh libgtpnl
+fi
 osmo-build-dep.sh libosmocore "" ac_cv_path_DOXYGEN=false
 
 verify_value_string_arrays_are_terminated.py $(find . -name "*.[hc]")
@@ -35,7 +38,7 @@ set -x
 
 cd "$base"
 autoreconf --install --force
-./configure
+./configure --enable-sanitize --enable-werror $GTP
 $MAKE $PARALLEL_MAKE
 $MAKE distcheck
 
