@@ -1,19 +1,19 @@
-/* 
+/*
  *  OsmoGGSN - Gateway GPRS Support Node
  *  Copyright (C) 2002, 2003, 2004 Mondru AB.
  *  Copyright (C) 2010-2011, 2016-2017 Harald Welte <laforge@gnumonks.org>
  *  Copyright (C) 2015-2017 sysmocom - s.f.m.c. GmbH
- * 
+ *
  *  The contents of this file may be used under the terms of the GNU
  *  General Public License Version 2, provided that the above copyright
  *  notice and this permission notice is included in all copies or
  *  substantial portions of the software.
- * 
+ *
  */
 
 /*
  * gtp.c: Contains all GTP functionality. Should be able to handle multiple
- * tunnels in the same program. 
+ * tunnels in the same program.
  *
  * TODO:
  *  - Do we need to handle fragmentation?
@@ -86,6 +86,51 @@ const char *gtp_version()
 	return VERSION;
 }
 
+const struct value_string gtp_type_names[] = {
+	{ GTP_ECHO_REQ,        "Echo Request" },
+	{ GTP_ECHO_RSP,        "Echo Response" },
+	{ GTP_NOT_SUPPORTED,   "Version Not Supported" },
+	{ GTP_ALIVE_REQ,       "Node Alive Request" },
+	{ GTP_ALIVE_RSP,       "Node Alive Response" },
+	{ GTP_REDIR_REQ,       "Redirection Request" },
+	{ GTP_REDIR_RSP,       "Redirection Response" },
+	{ GTP_CREATE_PDP_REQ,  "Create PDP Context Request" },
+	{ GTP_CREATE_PDP_RSP,  "Create PDP Context Response" },
+	{ GTP_UPDATE_PDP_REQ,  "Update PDP Context Request" },
+	{ GTP_UPDATE_PDP_RSP,  "Update PDP Context Response" },
+	{ GTP_DELETE_PDP_REQ,  "Delete PDP Context Request" },
+	{ GTP_DELETE_PDP_RSP,  "Delete PDP Context Response" },
+	{ GTP_ERROR,           "Error Indication" },
+	{ GTP_PDU_NOT_REQ,     "PDU Notification Request" },
+	{ GTP_PDU_NOT_RSP,     "PDU Notification Response" },
+	{ GTP_PDU_NOT_REJ_REQ, "PDU Notification Reject Request" },
+	{ GTP_PDU_NOT_REJ_RSP, "PDU Notification Reject Response" },
+	{ GTP_SUPP_EXT_HEADER, "Supported Extension Headers Notification" },
+	{ GTP_SND_ROUTE_REQ,   "Send Routeing Information for GPRS Request" },
+	{ GTP_SND_ROUTE_RSP,   "Send Routeing Information for GPRS Response" },
+	{ GTP_FAILURE_REQ,     "Failure Report Request" },
+	{ GTP_FAILURE_RSP,     "Failure Report Response" },
+	{ GTP_MS_PRESENT_REQ,  "Note MS GPRS Present Request" },
+	{ GTP_MS_PRESENT_RSP,  "Note MS GPRS Present Response" },
+	{ GTP_IDEN_REQ,        "Identification Request" },
+	{ GTP_IDEN_RSP,        "Identification Response" },
+	{ GTP_SGSN_CONTEXT_REQ,"SGSN Context Request" },
+	{ GTP_SGSN_CONTEXT_RSP,"SGSN Context Response" },
+	{ GTP_SGSN_CONTEXT_ACK,"SGSN Context Acknowledge" },
+	{ GTP_FWD_RELOC_REQ,   "Forward Relocation Request" },
+	{ GTP_FWD_RELOC_RSP,   "Forward Relocation Response" },
+	{ GTP_FWD_RELOC_COMPL, "Forward Relocation Complete" },
+	{ GTP_RELOC_CANCEL_REQ,"Relocation Cancel Request" },
+	{ GTP_RELOC_CANCEL_RSP,"Relocation Cancel Response" },
+	{ GTP_FWD_SRNS,        "Forward SRNS Context" },
+	{ GTP_FWD_RELOC_ACK,   "Forward Relocation Complete Acknowledge" },
+	{ GTP_FWD_SRNS_ACK,    "Forward SRNS Context Acknowledge" },
+	{ GTP_DATA_TRAN_REQ,   "Data Record Transfer Request" },
+	{ GTP_DATA_TRAN_RSP,   "Data Record Transfer Response" },
+	{ GTP_GPDU,            "G-PDU" },
+	{ 0, NULL }
+};
+
 /* gtp_new */
 /* gtp_free */
 
@@ -152,7 +197,7 @@ int gtp_set_cb_recovery(struct gsn_t *gsn,
 	return 0;
 }
 
-extern int gtp_set_cb_data_ind(struct gsn_t *gsn,
+int gtp_set_cb_data_ind(struct gsn_t *gsn,
 			       int (*cb_data_ind) (struct pdp_t * pdp,
 						   void *pack, unsigned len))
 {
@@ -168,7 +213,7 @@ extern int gtp_set_cb_data_ind(struct gsn_t *gsn,
  * to hold the packet header.
  * returns the length of the header. 0 on error.
  **/
-static unsigned int get_default_gtp(int version, uint8_t type, void *packet)
+static unsigned int get_default_gtp(uint8_t version, uint8_t type, void *packet)
 {
 	struct gtp0_header *gtp0_default = (struct gtp0_header *)packet;
 	struct gtp1_header_long *gtp1_default =
@@ -281,19 +326,19 @@ static uint32_t get_tei(void *pack)
 
 /* ***********************************************************
  * Reliable delivery of signalling messages
- * 
+ *
  * Sequence numbers are used for both signalling messages and
  * data messages.
  *
  * For data messages each tunnel maintains a sequence counter,
  * which is incremented by one each time a new data message
  * is sent. The sequence number starts at (0) zero at tunnel
- * establishment, and wraps around at 65535 (29.060 9.3.1.1 
+ * establishment, and wraps around at 65535 (29.060 9.3.1.1
  * and 09.60 8.1.1.1). The sequence numbers are either ignored,
  * or can be used to check the validity of the message in the
  * receiver, or for reordering af packets.
  *
- * For signalling messages the sequence number is used by 
+ * For signalling messages the sequence number is used by
  * signalling messages for which a response is defined. A response
  * message should copy the sequence from the corresponding request
  * message. The sequence number "unambiguously" identifies a request
@@ -311,7 +356,7 @@ static uint32_t get_tei(void *pack)
  * with path setup and teardown.
  *
  * If a response message is lost, the request will be retransmitted, and
- * the receiving GSN will receive a "duplicated" request. The standard 
+ * the receiving GSN will receive a "duplicated" request. The standard
  * requires the receiving GSN to send a response, with the same information
  * as in the original response. For most messages this happens automatically:
  *
@@ -326,22 +371,22 @@ static uint32_t get_tei(void *pack)
  *   a nonexist reply message.
  *
  * The correct solution will be to make a queue containing response messages.
- * This queue should be checked whenever a request is received. If the 
+ * This queue should be checked whenever a request is received. If the
  * response is allready in the queue that response should be transmitted.
  * It should be possible to find messages in this queue on the basis of
  * the sequence number and peer GSN IP address (The sequense number is unique
  * within each path). This need to be implemented by a hash table. Furthermore
  * it should be possibly to delete messages based on a timeout. This can be
  * achieved by means of a linked list. The timeout value need to be larger
- * than T3-RESPONSE * N3-REQUESTS (recommended value 5). These timers are 
+ * than T3-RESPONSE * N3-REQUESTS (recommended value 5). These timers are
  * set in the peer GSN, so there is no way to know these parameters. On the
  * other hand the timeout value need to be so small that we do not receive
  * wraparound sequence numbere before the message is deleted. 60 seconds is
  * probably not a bad choise.
- * 
+ *
  * This queue however is first really needed from gtp1.
  *
- * gtp_req: 
+ * gtp_req:
  *   Send off a signalling message with appropiate sequence
  *   number. Store packet in queue.
  * gtp_conf:
@@ -357,7 +402,7 @@ static uint32_t get_tei(void *pack)
  *   a predefined timeout.
  *************************************************************/
 
-int gtp_req(struct gsn_t *gsn, int version, struct pdp_t *pdp,
+static int gtp_req(struct gsn_t *gsn, uint8_t version, struct pdp_t *pdp,
 	    union gtp_packet *packet, int len,
 	    struct in_addr *inetaddr, void *cbp)
 {
@@ -432,7 +477,7 @@ int gtp_req(struct gsn_t *gsn, int version, struct pdp_t *pdp,
  * Remove signalling packet from retransmission queue.
  * return 0 on success, EOF if packet was not found */
 
-int gtp_conf(struct gsn_t *gsn, int version, struct sockaddr_in *peer,
+static int gtp_conf(struct gsn_t *gsn, uint8_t version, struct sockaddr_in *peer,
 	     union gtp_packet *packet, int len, uint8_t * type, void **cbp)
 {
 	uint8_t ver = GTPHDR_F_GET_VER(packet->flags);
@@ -523,7 +568,7 @@ int gtp_retranstimeout(struct gsn_t *gsn, struct timeval *timeout)
 	return 0;
 }
 
-int gtp_resp(int version, struct gsn_t *gsn, struct pdp_t *pdp,
+static int gtp_resp(uint8_t version, struct gsn_t *gsn, struct pdp_t *pdp,
 	     union gtp_packet *packet, int len,
 	     struct sockaddr_in *peer, int fd, uint16_t seq, uint64_t tid)
 {
@@ -581,7 +626,7 @@ int gtp_resp(int version, struct gsn_t *gsn, struct pdp_t *pdp,
 	return 0;
 }
 
-int gtp_notification(struct gsn_t *gsn, int version,
+static int gtp_notification(struct gsn_t *gsn, uint8_t version,
 		     union gtp_packet *packet, int len,
 		     struct sockaddr_in *peer, int fd, uint16_t seq)
 {
@@ -626,7 +671,7 @@ int gtp_notification(struct gsn_t *gsn, int version,
 	return 0;
 }
 
-int gtp_dublicate(struct gsn_t *gsn, int version,
+static int gtp_dublicate(struct gsn_t *gsn, uint8_t version,
 		  struct sockaddr_in *peer, uint16_t seq)
 {
 	struct qmsg_t *qmsg;
@@ -852,7 +897,7 @@ int gtp_free(struct gsn_t *gsn)
  * For response messages we need to be able to respond to
  * the relevant src port even if it is locally allocated by
  * the peer.
- * 
+ *
  * The need for path management!
  * We might need to keep a list of active paths. This might
  * be in the form of remote IP address + UDP port numbers.
@@ -942,9 +987,9 @@ int gtp_echo_conf(struct gsn_t *gsn, int version, struct sockaddr_in *peer,
 /* This message is somewhat special in that it actually is a
  * response to some other message with unsupported GTP version
  * For this reason it has parameters like a response, and does
- * its own message transmission. No signalling queue is used 
+ * its own message transmission. No signalling queue is used
  * The reply is sent to the peer IP and peer UDP. This means that
- * the peer will be receiving a GTP0 message on a GTP1 port! 
+ * the peer will be receiving a GTP0 message on a GTP1 port!
  * In practice however this will never happen as a GTP0 GSN will
  * only listen to the GTP0 port, and therefore will never receive
  * anything else than GTP0 */
@@ -971,7 +1016,7 @@ int gtp_unsup_ind(struct gsn_t *gsn, struct sockaddr_in *peer,
 }
 
 /* Send off an Supported Extension Headers Notification */
-int gtp_extheader_req(struct gsn_t *gsn, int version, struct sockaddr_in *peer,
+static int gtp_extheader_req(struct gsn_t *gsn, uint8_t version, struct sockaddr_in *peer,
 		      int fd, void *pack, unsigned len)
 {
 	union gtp_packet packet;
@@ -992,7 +1037,7 @@ int gtp_extheader_req(struct gsn_t *gsn, int version, struct sockaddr_in *peer,
 }
 
 /* Handle a Supported Extension Headers Notification */
-int gtp_extheader_ind(struct gsn_t *gsn, struct sockaddr_in *peer,
+static int gtp_extheader_ind(struct gsn_t *gsn, struct sockaddr_in *peer,
 		      void *pack, unsigned len)
 {
 
@@ -1007,7 +1052,7 @@ int gtp_extheader_ind(struct gsn_t *gsn, struct sockaddr_in *peer,
  * Messages: create, update and delete PDP context
  *
  * Information storage
- * Information storage for each PDP context is defined in 
+ * Information storage for each PDP context is defined in
  * 23.060 section 13.3. Includes IMSI, MSISDN, APN, PDP-type,
  * PDP-address (IP address), sequence numbers, charging ID.
  * For the SGSN it also includes radio related mobility
@@ -1015,7 +1060,7 @@ int gtp_extheader_ind(struct gsn_t *gsn, struct sockaddr_in *peer,
  *************************************************************/
 
 /* API: Send Create PDP Context Request (7.3.1) */
-extern int gtp_create_context_req(struct gsn_t *gsn, struct pdp_t *pdp,
+int gtp_create_context_req(struct gsn_t *gsn, struct pdp_t *pdp,
 				  void *cbp)
 {
 	union gtp_packet packet;
@@ -1093,7 +1138,7 @@ extern int gtp_create_context_req(struct gsn_t *gsn, struct pdp_t *pdp,
 				  pdp->cch_pdp);
 	}
 
-	/* TODO 
+	/* TODO
 	   gtpie_tv2(&packet, &length, GTP_MAX, GTPIE_TRACE_REF,
 	   pdp->traceref);
 	   gtpie_tv2(&packet, &length, GTP_MAX, GTPIE_TRACE_TYPE,
@@ -1537,7 +1582,7 @@ int gtp_create_pdp_ind(struct gsn_t *gsn, int version,
 		    (!memcmp(pdp->msisdn.v, pdp_old->msisdn.v, pdp->msisdn.l)))
 		{
 			/* OK! We are dealing with the same APN. We will copy new
-			 * parameters to the old pdp and send off confirmation 
+			 * parameters to the old pdp and send off confirmation
 			 * We ignore the following information elements:
 			 * QoS: MS will get originally negotiated QoS.
 			 * End user address (EUA). MS will get old EUA anyway.
@@ -1853,14 +1898,14 @@ int gtp_update_context(struct gsn_t *gsn, struct pdp_t *pdp, void *cbp,
 
 	gtpie_tv1(&packet, &length, GTP_MAX, GTPIE_NSAPI, pdp->nsapi);
 
-	/* TODO 
+	/* TODO
 	   gtpie_tv2(&packet, &length, GTP_MAX, GTPIE_TRACE_REF,
 	   pdp->traceref);
 	   gtpie_tv2(&packet, &length, GTP_MAX, GTPIE_TRACE_TYPE,
 	   pdp->tracetype); */
 
 	/* TODO if ggsn update message
-	   gtpie_tlv(&packet, &length, GTP_MAX, GTPIE_EUA, 
+	   gtpie_tlv(&packet, &length, GTP_MAX, GTPIE_EUA,
 	   pdp->eua.l, pdp->eua.v);
 	 */
 
@@ -1891,7 +1936,7 @@ int gtp_update_context(struct gsn_t *gsn, struct pdp_t *pdp, void *cbp,
 }
 
 /* Send Update PDP Context Response */
-int gtp_update_pdp_resp(struct gsn_t *gsn, int version,
+static int gtp_update_pdp_resp(struct gsn_t *gsn, uint8_t version,
 			struct sockaddr_in *peer, int fd,
 			void *pack, unsigned len,
 			struct pdp_t *pdp, uint8_t cause)
@@ -1932,8 +1977,8 @@ int gtp_update_pdp_resp(struct gsn_t *gsn, int version,
 		gtpie_tv4(&packet, &length, GTP_MAX, GTPIE_CHARGING_ID,
 			  pdp->teid_own);
 
-		/* If ggsn 
-		   gtpie_tlv(&packet, &length, GTP_MAX, GTPIE_EUA, 
+		/* If ggsn
+		   gtpie_tlv(&packet, &length, GTP_MAX, GTPIE_EUA,
 		   pdp->eua.l, pdp->eua.v); */
 
 		gtpie_tlv(&packet, &length, GTP_MAX, GTPIE_GSN_ADDR,
@@ -1953,7 +1998,7 @@ int gtp_update_pdp_resp(struct gsn_t *gsn, int version,
 }
 
 /* Handle Update PDP Context Request */
-int gtp_update_pdp_ind(struct gsn_t *gsn, int version,
+static int gtp_update_pdp_ind(struct gsn_t *gsn, uint8_t version,
 		       struct sockaddr_in *peer, int fd,
 		       void *pack, unsigned len)
 {
@@ -2127,7 +2172,7 @@ int gtp_update_pdp_ind(struct gsn_t *gsn, int version,
 	   GTP_LOGPKG(LOGL_ERROR, peer, pack, len,
 	   "Missing mandatory information field");
 	   memcpy(pdp, &pdp_backup, sizeof(pdp_backup));
-	   return gtp_update_pdp_resp(gsn, version, pdp, 
+	   return gtp_update_pdp_resp(gsn, version, pdp,
 	   GTPCAUSE_MAN_IE_MISSING);
 	   } */
 
@@ -2181,7 +2226,7 @@ int gtp_update_pdp_ind(struct gsn_t *gsn, int version,
 }
 
 /* Handle Update PDP Context Response */
-int gtp_update_pdp_conf(struct gsn_t *gsn, int version,
+static int gtp_update_pdp_conf(struct gsn_t *gsn, uint8_t version,
 			struct sockaddr_in *peer, void *pack, unsigned len)
 {
 	struct pdp_t *pdp;
@@ -2397,7 +2442,7 @@ int gtp_delete_pdp_resp(struct gsn_t *gsn, int version,
 					if (pdp_getgtp1
 					    (&secondary_pdp,
 					     linked_pdp->secondary_tei[n])) {
-						LOGP(DLGTP, LOGL_ERROR, 
+						LOGP(DLGTP, LOGL_ERROR,
 							"Unknown secondary PDP context\n");
 						return EOF;
 					}
@@ -2569,7 +2614,7 @@ int gtp_delete_pdp_conf(struct gsn_t *gsn, int version,
 }
 
 /* Send Error Indication (response to a GPDU message) - 3GPP TS 29.060 7.3.7 */
-int gtp_error_ind_resp(struct gsn_t *gsn, int version,
+static int gtp_error_ind_resp(struct gsn_t *gsn, uint8_t version,
 		       struct sockaddr_in *peer, int fd,
 		       void *pack, unsigned len)
 {
@@ -2591,7 +2636,7 @@ int gtp_error_ind_resp(struct gsn_t *gsn, int version,
 }
 
 /* Handle Error Indication */
-int gtp_error_ind_conf(struct gsn_t *gsn, int version,
+static int gtp_error_ind_conf(struct gsn_t *gsn, uint8_t version,
 		       struct sockaddr_in *peer, void *pack, unsigned len)
 {
 	union gtpie_member *ie[GTPIE_SIZE];
@@ -2635,22 +2680,28 @@ int gtp_error_ind_conf(struct gsn_t *gsn, int version,
 	GTP_LOGPKG(LOGL_ERROR, peer, pack, len,
 		    "Received Error Indication\n");
 
+	/* This is obvious from above code, given the semantics of the
+	 * functions above, but Coverity doesn't figure this out, so
+	 * let's make it clear. It's good style anyway in case above
+	 * code should ever change. */
+	OSMO_ASSERT(pdp);
+
 	if (gsn->cb_delete_context)
 		gsn->cb_delete_context(pdp);
 	pdp_freepdp(pdp);
 	return 0;
 }
 
-int gtp_gpdu_ind(struct gsn_t *gsn, int version,
+static int gtp_gpdu_ind(struct gsn_t *gsn, uint8_t version,
 		 struct sockaddr_in *peer, int fd, void *pack, unsigned len)
 {
 
-	int hlen = GTP1_HEADER_SIZE_SHORT;
+	int hlen;
 
-	/* Need to include code to verify packet src and dest addresses */
 	struct pdp_t *pdp;
 
-	if (version == 0) {
+	switch (version) {
+	case 0:
 		if (pdp_getgtp0
 		    (&pdp, ntoh16(((union gtp_packet *)pack)->gtp0.h.flow))) {
 			gsn->err_unknownpdp++;
@@ -2660,7 +2711,8 @@ int gtp_gpdu_ind(struct gsn_t *gsn, int version,
 						  len);
 		}
 		hlen = GTP0_HEADER_SIZE;
-	} else if (version == 1) {
+		break;
+	case 1:
 		if (pdp_getgtp1
 		    (&pdp, ntoh32(((union gtp_packet *)pack)->gtp1l.h.tei))) {
 			gsn->err_unknownpdp++;
@@ -2675,9 +2727,11 @@ int gtp_gpdu_ind(struct gsn_t *gsn, int version,
 			hlen = GTP1_HEADER_SIZE_LONG;
 		else
 			hlen = GTP1_HEADER_SIZE_SHORT;
-	} else {
+		break;
+	default:
 		GTP_LOGPKG(LOGL_ERROR, peer, pack, len,
 			    "Unknown version: %d\n", version);
+		return EOF;
 	}
 
 	/* If the GPDU was not from the peer GSN tell him to delete context */
@@ -2694,10 +2748,10 @@ int gtp_gpdu_ind(struct gsn_t *gsn, int version,
 	return 0;
 }
 
-/* Receives GTP packet and sends off for further processing 
+/* Receives GTP packet and sends off for further processing
  * Function will check the validity of the header. If the header
- * is not valid the packet is either dropped or a version not 
- * supported is returned to the peer. 
+ * is not valid the packet is either dropped or a version not
+ * supported is returned to the peer.
  * TODO: Need to decide on return values! */
 int gtp_decaps0(struct gsn_t *gsn)
 {
@@ -2706,7 +2760,7 @@ int gtp_decaps0(struct gsn_t *gsn)
 	socklen_t peerlen;
 	int status;
 	struct gtp0_header *pheader;
-	int version = 0;	/* GTP version should be determined from header! */
+	uint8_t version;
 	int fd = gsn->fd0;
 
 	/* TODO: Need strategy of userspace buffering and blocking */
@@ -2742,15 +2796,17 @@ int gtp_decaps0(struct gsn_t *gsn)
 
 		pheader = (struct gtp0_header *)(buffer);
 
+		version = GTPHDR_F_GET_VER(pheader->flags);
+
 		/* Version should be gtp0 (or earlier) */
 		/* 09.60 is somewhat unclear on this issue. On gsn->fd0 we expect only */
 		/* GTP 0 messages. If other version message is received we reply that we */
 		/* only support version 0, implying that this is the only version */
 		/* supported on this port */
-		if (GTPHDR_F_GET_VER(pheader->flags) > 0) {
+		if (version > 0) {
 			gsn->unsup++;
-			GTP_LOGPKG(LOGL_ERROR, &peer, buffer,
-				    status, "Unsupported GTP version\n");
+			GTP_LOGPKG(LOGL_ERROR, &peer, buffer, status,
+				"Unsupported GTP version %"PRIu8"\n", version);
 			gtp_unsup_req(gsn, 0, &peer, gsn->fd0, buffer, status);	/* 29.60: 11.1.1 */
 			continue;
 		}
@@ -2851,7 +2907,7 @@ int gtp_decaps1c(struct gsn_t *gsn)
 	socklen_t peerlen;
 	int status;
 	struct gtp1_header_short *pheader;
-	int version = 1;	/* TODO GTP version should be determined from header! */
+	uint8_t version;
 	int fd = gsn->fd1c;
 
 	/* TODO: Need strategy of userspace buffering and blocking */
@@ -2887,11 +2943,13 @@ int gtp_decaps1c(struct gsn_t *gsn)
 
 		pheader = (struct gtp1_header_short *)(buffer);
 
+		version = GTPHDR_F_GET_VER(pheader->flags);
+
 		/* Version must be no larger than GTP 1 */
-		if (GTPHDR_F_GET_VER(pheader->flags) > 1) {
+		if (version > 1) {
 			gsn->unsup++;
-			GTP_LOGPKG(LOGL_ERROR, &peer, buffer,
-				    status, "Unsupported GTP version\n");
+			GTP_LOGPKG(LOGL_ERROR, &peer, buffer, status,
+				"Unsupported GTP version %"PRIu8"\n", version);
 			gtp_unsup_req(gsn, version, &peer, fd, buffer, status);
 			/*29.60: 11.1.1 */
 			continue;
@@ -2901,10 +2959,10 @@ int gtp_decaps1c(struct gsn_t *gsn)
 		/* 29.060 is somewhat unclear on this issue. On gsn->fd1c we expect only */
 		/* GTP 1 messages. If GTP 0 message is received we silently discard */
 		/* the message */
-		if (GTPHDR_F_GET_VER(pheader->flags) < 1) {
+		if (version < 1) {
 			gsn->unsup++;
-			GTP_LOGPKG(LOGL_ERROR, &peer, buffer,
-				    status, "Unsupported GTP version\n");
+			GTP_LOGPKG(LOGL_ERROR, &peer, buffer, status,
+				"Unsupported GTP version %"PRIu8"\n", version);
 			continue;
 		}
 
@@ -3026,7 +3084,7 @@ int gtp_decaps1u(struct gsn_t *gsn)
 	socklen_t peerlen;
 	int status;
 	struct gtp1_header_short *pheader;
-	int version = 1;	/* GTP version should be determined from header! */
+	uint8_t version;
 	int fd = gsn->fd1u;
 
 	/* TODO: Need strategy of userspace buffering and blocking */
@@ -3063,11 +3121,13 @@ int gtp_decaps1u(struct gsn_t *gsn)
 
 		pheader = (struct gtp1_header_short *)(buffer);
 
+		version = GTPHDR_F_GET_VER(pheader->flags);
+
 		/* Version must be no larger than GTP 1 */
-		if (GTPHDR_F_GET_VER(pheader->flags) > 1) {
+		if (version > 1) {
 			gsn->unsup++;
-			GTP_LOGPKG(LOGL_ERROR, &peer, buffer,
-				    status, "Unsupported GTP version\n");
+			GTP_LOGPKG(LOGL_ERROR, &peer, buffer, status,
+				"Unsupported GTP version %"PRIu8"\n", version);
 			gtp_unsup_req(gsn, 1, &peer, gsn->fd1c, buffer, status);	/*29.60: 11.1.1 */
 			continue;
 		}
@@ -3076,10 +3136,10 @@ int gtp_decaps1u(struct gsn_t *gsn)
 		/* 29.060 is somewhat unclear on this issue. On gsn->fd1c we expect only */
 		/* GTP 1 messages. If GTP 0 message is received we silently discard */
 		/* the message */
-		if (GTPHDR_F_GET_VER(pheader->flags) < 1) {
+		if (version < 1) {
 			gsn->unsup++;
-			GTP_LOGPKG(LOGL_ERROR, &peer, buffer,
-				    status, "Unsupported GTP version\n");
+			GTP_LOGPKG(LOGL_ERROR, &peer, buffer, status,
+				"Unsupported GTP version %"PRIu8"\n", version);
 			continue;
 		}
 
@@ -3235,19 +3295,10 @@ int gtp_data_req(struct gsn_t *gsn, struct pdp_t *pdp, void *pack, unsigned len)
  * Conversion functions
  *************************************************************/
 
-int char2ul_t(char *src, struct ul_t dst)
-{
-	dst.l = strlen(src) + 1;
-	dst.v = malloc(dst.l);
-	dst.v[0] = dst.l - 1;
-	memcpy(&dst.v[1], src, dst.v[0]);
-	return 0;
-}
-
 /* ***********************************************************
  * IP address conversion functions
  * There exist several types of address representations:
- * - eua: End User Address. (29.060, 7.7.27, message type 128) 
+ * - eua: End User Address. (29.060, 7.7.27, message type 128)
  *   Used for signalling address to mobile station. Supports IPv4
  *   IPv6 x.25 etc. etc.
  * - gsna: GSN Address. (29.060, 7.7.32, message type 133): IP address
@@ -3301,7 +3352,7 @@ int in_addr2gsna(struct ul16_t *gsna, struct in_addr *src)
  * _network byte order_ to contain BCD digits ?!? */
 const char *imsi_gtp2str(const uint64_t *imsi)
 {
-	static char buf[sizeof(*imsi)+1];
+	static char buf[sizeof(*imsi)*2+1];
 	const uint8_t *imsi8 = (const uint8_t *) imsi;
 	unsigned int i, j = 0;
 
